/* File: alchemy.c */

/* Purpose: Alchemy initializing code */

#include "angband.h"


/* Array of k_idxs of all potions, by sval */

static int potion_list[SV_POTION_MAX];



/* Initialize alchemy stuff */

void alchemy_init(void)
{
        int     i, j;           /* svals of mixed potions */
        int     k1, k2;         /* k_idx of i and j */
        int     sv;             /* sval of product */
        int     kprod;          /* k_idx of product */
        int     attempts;

        object_kind *k_ptr;

        /* Hack: use "special seed" to get consistent formulae */
        Rand_quick = TRUE;
        Rand_value = seed_alchemy;

        /* Build list of legal potions */
        for (sv = 0; sv < SV_POTION_MAX; sv++)
        {
                potion_list[sv] = -1;

                for (j = 0; j < max_k_idx; j++)
                {
                        k_ptr = &k_info[j];
                        if ((k_ptr->tval == TV_POTION) && (k_ptr->sval == sv))
                        {
                                potion_list[sv] = j;
                        }
                }
        }

        for (i=0; i < SV_POTION_MAX; i++)
        {
                for (j=0; j < i; j++)
                {
                        k1 = potion_list[i];
                        k2 = potion_list[j];

                        if ((k1 == -1) || (k2 == -1))
                        {
                                potion_alch[i][j] = -1;
                                continue;
                        }

                        attempts = 0;

                        /* Try to find a suitable product for the pair */
                        while (TRUE)
                        {
                                sv = rand_int(SV_POTION_MAX);

                                /* Neither can be the resulting potion */
                                if ((i == sv) || (j == sv)) continue;

                                kprod = potion_list[sv];

                                /* Must be a legal potion */
                                if (kprod == -1) continue;

                                attempts++;

                                k_ptr = &k_info[kprod];

                                /* Expensive potions must be made from good potions */
                                if ((k_ptr->cost > 500) &&
                                        (k_ptr->cost > k_info[k1].cost * k_info[k2].cost))
                                        continue;

                                /* Prefer if the cost is more than the sum of the parts */
                                if (k_info[k1].cost + k_info[k2].cost > k_ptr->cost)
                                {
                                        /* Up to 20 failures allowed */
                                        if (attempts < 20) continue;
                                }

                                /* We've found a good potion... stop. */
                                break;
                        }

                        potion_alch[i][j] = kprod;
                }
        }

        /* Mirror-reflect the table so "A and B" is the same as "B and A" */

        for (i=0; i < SV_POTION_MAX; i++)
        {
                for (j=0; j < i; j++)
                {
                        potion_alch[j][i] = potion_alch[i][j];
                }

                /* Paranoia */
                potion_alch[i][i] = -1;
        }

        Rand_quick = FALSE;
        return;
}


/* Describe a potion (as object_desc()) from its k_idx. */
static void potion_desc(char *buf, int k_idx)
{
        object_type forge;

        object_wipe(&forge);
        object_prep(&forge, k_idx);
        object_desc(buf, &forge, 0, 0);
        return;
}

/*
 * Display the formulae involving this potion.
 */
void alchemy_roff(int k_idx)
{
        int i, j;
        int o_idx;
        object_kind *k_ptr = &k_info[k_idx];
        int sv;

        char ibuf[80], jbuf[80];
        char obuf[80];

        char info[128][100];
        int  l = 0;
        int  x, y;

        bool found;

        if (k_ptr->tval != TV_POTION) return;

        sv = k_ptr->sval;

        /* Get name of potion */
        potion_desc(obuf, k_idx);

        found = FALSE;

        /* First, display what we know about how to make it */
        for (i=0; i < SV_POTION_MAX; i++)
        {
                if (potion_list[i] == -1) continue;

                for (j=0; j < i; j++)
                {
                        if (potion_list[j] == -1) continue;

                        if ((cheat_know || r_potion_alch[i][j]) &&
                                (potion_alch[i][j] == k_idx))
                        {
                                potion_desc(ibuf, potion_list[i]);
                                potion_desc(jbuf, potion_list[j]);

                                strcpy(info[l++], format("%s + %s = %s", ibuf, jbuf, obuf));
                                found = TRUE;
                        }
                }
        }

        /* Next, display what we know about what it makes */
        for (i=0; i < SV_POTION_MAX; i++)
        {
                if (potion_list[i] == -1) continue;

                if ((cheat_know) || r_potion_alch[sv][i])
                {
                        potion_desc(ibuf, potion_list[i]);
                        o_idx = potion_alch[sv][i];
                        if (o_idx != -1)
                        {
                                potion_desc(jbuf, o_idx);
                        }
                        else
                        {
                                strcpy(jbuf, "nothing");
                        }

                        strcpy(info[l++], format("%s + %s = %s", obuf, ibuf, jbuf));
                        found = TRUE;
                }
        }

        if (!found)
        {
                strcpy(info[l++], "You don't know anything about this potion.");
        }

        /* Save the screen */
        screen_save();

        /* Flush messages */
        msg_print(NULL);

        /* Label the information */
        prt("     Alchemical Information:", 0, 0);
        roff("\n\n");

        /* We will print on top of the map, using roff() */
        for (j = 0; j < l; j++)
        {
                /* Show the info */
                roff(info[j]);
                roff("\n");

                (void)Term_locate(&x, &y);

                /* Every 10 entries (lines 2 to ~11), start over */
                if ((y >= 12) && (j+1 < l))
                {
                        roff("-- more --");
                        inkey();
                        screen_load();
                        screen_save();
                        msg_print(NULL);

                        /* Label the information again */
                        prt("     Alchemical Information:", 0, 0);
                        roff("\n\n");
                }
        }

        /* Wait for it */
        roff("[Press any key to continue]");
        inkey();

        /* Restore the screen */
        screen_load();
}

void alchemy_id(object_type *o_ptr)
{
        /*
         * *ID*'ing a potion recalls about 5 of the formulae it is
         * involved in (and which involve only potions you have tasted).
         */

        int i, k;
        object_kind *k_ptr;

        int realized = 0;

        if (o_ptr->tval != TV_POTION) return;

        for (i=0; (realized < 5) && (i < SV_POTION_MAX); i++)
        {
                if (i == o_ptr->sval) continue;
                if (potion_list[i] == -1) continue;

                k = potion_alch[i][o_ptr->sval];
                if (k >= 0)
                {
                        k_ptr = &k_info[k];
                        if (((k_ptr->aware) || (k_ptr->tried)) && one_in_(2))
                        {
                                r_potion_alch[i][o_ptr->sval] = TRUE;
                                r_potion_alch[o_ptr->sval][i] = TRUE;
                                realized++;
                        }
                }

                k_ptr = &k_info[potion_list[i]];
                if ((k_ptr->aware) || (k_ptr->tried))
                {
                        for (k=0; k < SV_POTION_MAX; k++)
                        {
                                if (k == o_ptr->sval) continue;
                                if (potion_list[k] == -1) continue;

                                if (potion_alch[i][k] == o_ptr->k_idx)
                                {
                                        k_ptr = &k_info[potion_list[k]];
                                        if (((k_ptr->aware) || (k_ptr->tried))
                                            && one_in_(2))
                                        {
                                                r_potion_alch[i][k] = TRUE;
                                                realized++;
                                                if (realized >= 5) break;
                                        }
                                }
                        }
                }
                /* end of for (i) loop */
        }
        return;
}


