/* File: cmd-know.c */

/* Purpose: Monster and object knowledge sub-commands */

/*
 * Copyright (c) 1997 Ben Harrison, James E. Wilson, Robert A. Koeneke
 *
 * This software may be copied and distributed for educational, research,
 * and not for profit purposes provided that this copyright and statement
 * are included in all such copies.  Other copyrights may also apply.
 */

#include "angband.h"


#define BROWSER_ROWS    16

/* Length of malloc'ed "monster list" arrays */
#define M_LIST_ITEMS    (max_r_idx + 1)

/*
 * Structure for building monster "lists"
 */
typedef struct monster_list_entry monster_list_entry;

struct monster_list_entry
{
        s16b r_idx;                     /* Monster race index */
        s16b u_idx;                     /* Unique index (unused) */

        byte amount;
};

/*
 * Description of each monster group.
 */
static cptr monster_group_text[] =
{
        "  Uniques",
        "A Angels",
        "a Ants",
        "B Birds",
        "b Bats",
        "C Canines",
        "c Centipedes",
        "  Dragons",
        "E Elementals",
        "e Eyes/Beholders",
        "F Flies",
        "f Felines",
        "G Ghosts",
        "g Golems",
        "H Hybrids",
        "h Humanoids",
        "I Insects",
        "i Icky Things",
        "J Snakes",
        "j Jellies",
        "K Killer Beetles",
        "k Kobolds",
        "l Lice",
        "L Lichs",
        "  Mimics",
        "M Multi-headed Hydras",
        "m Molds",
        ", Mushroom Patches",
        "n Nagas",
        "O Ogres",
        "o Orcs",
        "P Giants",
        "p Humans",
        "Q Quylthulgs",
        "q Quadropeds",
        "R Reptiles/Amphibians",
        "r Rodents",
        "S Scorpions/Spiders",
        "s Skeletons",
        "T Trolls",
        "t Townspeople",
        "  Demons",
        "V Vampires",
        "v Vortices",
        "W Wights/Wraiths",
        "w Worms/Worm Masses",
        "X Xorns/Xarens",
        "x Extraterrestrials",
        "Y Yeti",
        "y Yeeks",
        "Z Zephyr Hounds",
        "z Lesser Undead",
        "  All Monsters",
        NULL
};

/*
 * Symbols of monsters in each group. Note the "Uniques" group
 * is handled differently.
 */
static cptr monster_group_char[] =
{
        (char *) -1L,
        "A",
        "a",
        "B",
        "b",
        "C",
        "c",
        "Dd",
        "E",
        "e",
        "F",
        "f",
        "G",
        "g",
        "H",
        "h",
        "I",
        "i",
        "J",
        "j",
        "K",
        "k",
        "L",
        "l",
        "@$!?=.|~[]%",
        "M",
        "m",
        ",",
        "n",
        "O",
        "o",
        "P",
        "p",
        "Q",
        "q",
        "R",
        "r",
        "S",
        "s",
        "T",
        "t",
        "Uu",
        "V",
        "v",
        "W",
        "w",
        "X",
        "x",
        "Y",
        "y",
        "Z",
        "z",
        "AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz,@$!?=.|~[]%",
        NULL
};


/*
 * Build a list of monster indexes in the given group. Return the number
 * of monsters in the group.
 */
static int collect_monsters(int grp_cur, monster_list_entry *mon_idx, int mode)
{
        int i, mon_cnt = 0;

        /* Get a list of x_char in this group */
        cptr group_char = monster_group_char[grp_cur];

        /* XXX Hack -- Check if this is the "Uniques" group */
        bool grp_unique = (monster_group_char[grp_cur] == (char *) -1L);

        /* Check every race */
        for (i = 0; i < max_r_idx; i++)
        {
                /* Access the race */
                monster_race *r_ptr = &r_info[i];

                /* Is this a unique? */
                bool unique = ((r_ptr->flags1 & RF1_UNIQUE) ? TRUE : FALSE);

                if (r_ptr->flags3 & RF3_SEMI) unique = TRUE;

                /* Skip empty race */
                if (!r_ptr->name) continue;

                if (grp_unique && !(unique)) continue;

                /* Require known monsters */
                if (!(cheat_know) && !(r_ptr->r_sights)) continue;

                /* Check for race in the group */
                if (grp_unique || strchr(group_char, r_ptr->d_char))
                {
                        /* Add the race */
                        mon_idx[mon_cnt++].r_idx = i;

                        /* XXX Hack -- Just checking for non-empty group */
                        if (mode & 0x01) break;
                }
        }

        /* Terminate the list */
        mon_idx[mon_cnt++].r_idx = -1;
        mon_idx[mon_cnt].r_idx = 0;

        /* Return the number of races */
        return mon_cnt;
}

/*
 * Description of each object group.
 */
static cptr object_group_text[] =
{
        "Mushrooms",
        "Potions",
        "Amulets",
        "Rings",
        "Bows and Firearms",
        "Digging Tools",
        "Hafted Weapons",
        "Polearms",
        "Swords",
        "Boots",
        "Gloves",
        "Helms",
        "Crowns",
        "Shields",
        "Cloaks",
        "Soft Armor",
        "Hard Armor",
        "Dragon Scale Mail",
        "Lights",
        "Staves",
        "Wands",
        "Rods",
        "Scrolls",
        "Life Spellbooks",
        "Sorcery Spellbooks",
        "Nature Spellbooks",
        "Arcane Spellbooks",
        "Chaos Spellbooks",
        "Death Spellbooks",
        "Trump Spellbooks",
        NULL
};

/*
 * TVALs of items in each group
 */
static byte object_group_tval[] =
{
        TV_FOOD,
        TV_POTION,
        TV_AMULET,
        TV_RING,
        TV_BOW,
        TV_DIGGING,
        TV_HAFTED,
        TV_POLEARM,
        TV_SWORD,
        TV_BOOTS,
        TV_GLOVES,
        TV_HELM,
        TV_CROWN,
        TV_SHIELD,
        TV_CLOAK,
        TV_SOFT_ARMOR,
        TV_HARD_ARMOR,
        TV_DRAG_ARMOR,
        TV_LITE,
        TV_STAFF,
        TV_WAND,
        TV_ROD,
        TV_SCROLL,
        TV_LIFE_BOOK,
        TV_SORCERY_BOOK,
        TV_NATURE_BOOK,
        TV_ARCANE_BOOK,
        TV_CHAOS_BOOK,
        TV_DEATH_BOOK,
        TV_TRUMP_BOOK,
        0
};

/*
 * Build a list of object indexes in the given group. Return the number
 * of objects in the group.
 */
static int collect_objects(int grp_cur, int object_idx[])
{
        int i, j, k, object_cnt = 0;

        /* Get a list of x_char in this group */
        byte group_tval = object_group_tval[grp_cur];

        /* Check every object */
        for (i = 0; i < max_k_idx; i++)
        {
                /* Access the race */
                object_kind *k_ptr = &k_info[i];

                /* Skip empty objects */
                if (!k_ptr->name) continue;

                /* Skip non-flavoured objects */
                if (!k_ptr->flavor) continue;

                /* Skip items with no distribution (special artifacts) */
                for (j = 0, k = 0; j < 4; j++) k += k_ptr->chance[j];
                if (!(k))  continue;

                /* Require objects to be "aware" */
                if (!(k_ptr->aware)) continue;

                /* Check for race in the group */
                if (k_ptr->tval == group_tval)
                {
                        /* Add the race */
                        object_idx[object_cnt++] = i;
                }
        }

        /* Terminate the list */
        object_idx[object_cnt] = 0;

        /* Return the number of races */
        return object_cnt;
}

/*
 * Display the object groups.
 */
static void display_group_list(int col, int row, int wid, int per_page,
        int grp_idx[], cptr group_text[], int grp_cur, int grp_top)
{
        int i;

        /* Display lines until done */
        for (i = 0; i < per_page && (grp_idx[i] >= 0); i++)
        {
                /* Get the group index */
                int grp = grp_idx[grp_top + i];

                /* Choose a color */
                byte attr = (grp_top + i == grp_cur) ? TERM_L_BLUE : TERM_WHITE;

                /* Erase the entire line */
                Term_erase(col, row + i, wid);

                /* Display the group label */
                c_put_str(attr, group_text[grp], row + i, col);
        }
}

/*
 * Move the cursor in a browser window
 */
static void browser_cursor(char ch, int *column, int *grp_cur, int grp_cnt,
                                                   int *list_cur, int list_cnt)
{
        int d;
        int col = *column;
        int grp = *grp_cur;
        int list = *list_cur;

        /* Extract direction */
        d = get_keymap_dir(ch);

        if (!d) return;

        /* Diagonals - hack */
        if ((ddx[d] > 0) && ddy[d])
        {
                /* Browse group list */
                if (!col)
                {
                        int old_grp = grp;

                        /* Move up or down */
                        grp += ddy[d] * BROWSER_ROWS;

                        /* Verify */
                        if (grp >= grp_cnt)     grp = grp_cnt - 1;
                        if (grp < 0) grp = 0;
                        if (grp != old_grp)     list = 0;
                }

                /* Browse sub-list list */
                else
                {
                        /* Move up or down */
                        list += ddy[d] * BROWSER_ROWS;

                        /* Verify */
                        if (list >= list_cnt) list = list_cnt - 1;
                        if (list < 0) list = 0;
                }

                (*grp_cur) = grp;
                (*list_cur) = list;

                return;
        }

        if (ddx[d])
        {
                col += ddx[d];
                if (col < 0) col = 0;
                if (col > 1) col = 1;

                (*column) = col;

                return;
        }

        /* Browse group list */
        if (!col)
        {
                int old_grp = grp;

                /* Move up or down */
                grp += ddy[d];

                /* Verify */
                if (grp < 0) grp = 0;
                if (grp >= grp_cnt)     grp = grp_cnt - 1;
                if (grp != old_grp)     list = 0;
        }

        /* Browse sub-list list */
        else
        {
                /* Move up or down */
                list += ddy[d];

                /* Verify */
                if (list < 0) list = 0;
                if (list >= list_cnt) list = list_cnt - 1;
        }

        (*grp_cur) = grp;
        (*list_cur) = list;
}

/*
 * Display the monsters in a group.
 */
static void display_monster_list(int col, int row, int per_page, monster_list_entry *mon_idx,
        int mon_cur, int mon_top)
{
        int i;

        /* Display lines until done */
        for (i = 0; i < per_page && mon_idx[i].r_idx; i++)
        {
                byte attr;
                int r_idx;

                /* Choose a color */
                attr = ((i + mon_top == mon_cur) ? TERM_L_BLUE : TERM_WHITE);

                /* Get the monster race */
                r_idx = mon_idx[mon_top + i].r_idx;

                if (r_idx == -1)
                {
                        int t;
                        int totalkills = 0;

                        for (t = 0; mon_idx[t].r_idx > 0; t++)
                        {
                                totalkills += r_info[mon_idx[t].r_idx].r_pkills;
                        }

                        c_prt(attr, "***Total***", row + i, col);
                        put_str(format("%5d", totalkills), row + i, 73);
                }
                else
                {
                        /* Access the race */
                        monster_race *r_ptr = &r_info[r_idx];

                        /* Display the name */
                        c_prt(attr, format("%^s", (r_name + r_ptr->name)), row + i, col);

                        if (p_ptr->wizard)
                        {
                                c_prt(attr, format ("%d", r_idx), row + i, 60);
                        }

                        /* Display symbol */
                        Term_putch(68, row + i, r_ptr->x_attr, r_ptr->x_char);

                        /* Display kills */
                        if (r_ptr->flags1 & (RF1_UNIQUE))
                        {
                                put_str(format("%s", (r_ptr->max_num == 0) ? " dead" :
                                        "alive"), row + i, 73);
                        }
                        else if (r_ptr->flags3 & (RF3_SEMI))
                        {
                                if (r_ptr->max_num == 0)
                                        put_str(" dead", row + i, 73);
                                else
                                        put_str(format("%5d", r_ptr->r_pkills), row + i, 73);
                        }
                        else
                        {
                                put_str(format("%5d", r_ptr->r_pkills), row + i, 73);
                        }
                }
        }

        /* Clear remaining lines */
        for (; i < per_page; i++)
        {
                Term_erase(col, row + i, 255);
        }
}

/*
 * Display known monsters.
 */
void do_cmd_knowledge_monsters(void)
{
        int i, len, max;
        int grp_cur, grp_top;
        int mon_cur, mon_top;
        int grp_cnt, grp_idx[100];
        int mon_cnt;
        monster_list_entry *mon_idx;

        int column = 0;
        bool flag;
        bool redraw;

        /* Allocate the "mon_idx" array */
        C_MAKE(mon_idx, M_LIST_ITEMS, monster_list_entry);

        max = 0;
        grp_cnt = 0;

        /* Check every group */
        for (i = 0; monster_group_text[i] != NULL; i++)
        {
                /* Measure the label */
                len = strlen(monster_group_text[i]);

                /* Save the maximum length */
                if (len > max) max = len;

                /* See if any monsters are known */
                if ((monster_group_char[i] == ((char *) -1L)) ||
                        collect_monsters(i, mon_idx, 0x01))
                {
                        /* Build a list of groups with known monsters */
                        grp_idx[grp_cnt++] = i;
                }
        }

        /* Terminate the list */
        grp_idx[grp_cnt] = -1;

        grp_cur = grp_top = 0;
        mon_cur = mon_top = 0;

        flag = FALSE;
        redraw = TRUE;

        while (!flag)
        {
                char ch;

                if (redraw)
                {
                        clear_from(0);

                        prt("Knowledge - Monsters", 2, 0);
                        prt("Group", 4, 0);
                        prt("Name", 4, max + 3);
                        if (p_ptr->wizard) prt("Idx", 4, 60);
                        prt("Sym   Kills", 4, 67);

                        for (i = 0; i < 78; i++)
                        {
                                Term_putch(i, 5, TERM_WHITE, '=');
                        }

                        for (i = 0; i < BROWSER_ROWS; i++)
                        {
                                Term_putch(max + 1, 6 + i, TERM_WHITE, '|');
                        }

                        redraw = FALSE;
                }

                /* Scroll group list */
                if (grp_cur < grp_top) grp_top = grp_cur;
                if (grp_cur >= grp_top + BROWSER_ROWS) grp_top = grp_cur - BROWSER_ROWS + 1;

                /* Scroll monster list */
                if (mon_cur < mon_top) mon_top = mon_cur;
                if (mon_cur >= mon_top + BROWSER_ROWS) mon_top = mon_cur - BROWSER_ROWS + 1;

                /* Display a list of monster groups */
                display_group_list(0, 6, max, BROWSER_ROWS, grp_idx, monster_group_text, grp_cur, grp_top);

                /* Get a list of monsters in the current group */
                mon_cnt = collect_monsters(grp_idx[grp_cur], mon_idx, 0x00);

                /* Display a list of monsters in the current group */
                display_monster_list(max + 3, 6, BROWSER_ROWS, mon_idx, mon_cur, mon_top);

                /* Prompt */
                prt("<dir>, 'r' to recall, ESC", 23, 0);

                /* Mega Hack -- track this monster race */
                monster_race_track(mon_idx[mon_cur].r_idx);

                /* Hack -- handle stuff */
                handle_stuff();

                if (!column)
                {
                        Term_gotoxy(0, 6 + (grp_cur - grp_top));
                }
                else
                {
                        Term_gotoxy(max + 3, 6 + (mon_cur - mon_top));
                }

                ch = inkey();

                switch (ch)
                {
                        case ESCAPE:
                        {
                                flag = TRUE;
                                break;
                        }

                        case 'R':
                        case 'r':
                        {
                                /* Recall on screen */
                                if (mon_idx[mon_cur].r_idx > 0)
                                {
                                        screen_roff(mon_idx[mon_cur].r_idx, 0);

                                        (void) inkey();

                                        redraw = TRUE;
                                }
                                break;
                        }

                        default:
                        {
                                /* Move the cursor */
                                browser_cursor(ch, &column, &grp_cur, grp_cnt, &mon_cur, mon_cnt);

                                break;
                        }
                }
        }

        /* XXX XXX Free the "mon_idx" array */
        C_KILL(mon_idx, M_LIST_ITEMS, monster_list_entry);
}

/*
 * Display the objects in a group.
 */
static void display_object_list(int col, int row, int per_page, int object_idx[],
        int object_cur, int object_top)
{
        int i;

        /* Display lines until done */
        for (i = 0; i < per_page && object_idx[i]; i++)
        {
                /* Get the object index */
                int k_idx = object_idx[object_top + i];

                /* Access the object */
                object_kind *k_ptr = &k_info[k_idx];

                /* Choose a color */
                byte attr = ((k_ptr->aware) ? TERM_WHITE : TERM_SLATE);
                byte cursor = ((k_ptr->aware) ? TERM_L_BLUE : TERM_BLUE);
                attr = ((i + object_top == object_cur) ? cursor : attr);

                /* Display the name */
                c_prt(attr, k_name + k_ptr->name, row + i, col);

                if (p_ptr->wizard) c_prt(attr, format ("%d", k_idx), row + i, 70);

                if (k_ptr->aware)
                {
                        byte a = misc_to_attr[k_ptr->flavor];
                        byte c = misc_to_char[k_ptr->flavor];

                        /* Display symbol */
                        Term_putch(76, row + i, a, c);
                }
        }

        /* Clear remaining lines */
        for (; i < per_page; i++)
        {
                Term_erase(col, row + i, 255);
        }
}

/*
 * Display known objects
 */
void do_cmd_knowledge_objects(void)
{
        int i, len, max;
        int grp_cur, grp_top;
        int object_old, object_cur, object_top;
        int grp_cnt, grp_idx[100];
        int object_cnt;
        int *object_idx;

        int column = 0;
        bool flag;
        bool redraw;

        /* Allocate the "object_idx" array */
        C_MAKE(object_idx, max_k_idx, int);

        max = 0;
        grp_cnt = 0;

        /* Check every group */
        for (i = 0; object_group_text[i] != NULL; i++)
        {
                /* Measure the label */
                len = strlen(object_group_text[i]);

                /* Save the maximum length */
                if (len > max) max = len;

                /* See if any monsters are known */
                if (collect_objects(i, object_idx))
                {
                        /* Build a list of groups with known monsters */
                        grp_idx[grp_cnt++] = i;
                }
        }

        /* Terminate the list */
        grp_idx[grp_cnt] = -1;

        grp_cur = grp_top = 0;
        object_cur = object_top = 0;
        object_old = -1;

        flag = FALSE;
        redraw = TRUE;

        while (!flag)
        {
                char ch;

                if (redraw)
                {
                        clear_from(0);

                        prt("Knowledge - objects", 2, 0);
                        prt("Group", 4, 0);
                        prt("Name", 4, max + 3);
                        if (p_ptr->wizard) prt("Idx", 4, 70);
                        prt("Sym", 4, 75);

                        for (i = 0; i < 78; i++)
                        {
                                Term_putch(i, 5, TERM_WHITE, '=');
                        }

                        for (i = 0; i < BROWSER_ROWS; i++)
                        {
                                Term_putch(max + 1, 6 + i, TERM_WHITE, '|');
                        }

                        redraw = FALSE;
                }

                /* Scroll group list */
                if (grp_cur < grp_top) grp_top = grp_cur;
                if (grp_cur >= grp_top + BROWSER_ROWS) grp_top = grp_cur - BROWSER_ROWS + 1;

                /* Scroll monster list */
                if (object_cur < object_top) object_top = object_cur;
                if (object_cur >= object_top + BROWSER_ROWS) object_top = object_cur - BROWSER_ROWS + 1;

                /* Display a list of object groups */
                display_group_list(0, 6, max, BROWSER_ROWS, grp_idx, object_group_text, grp_cur, grp_top);

                /* Get a list of objects in the current group */
                object_cnt = collect_objects(grp_idx[grp_cur], object_idx);

                /* Display a list of objects in the current group */
                display_object_list(max + 3, 6, BROWSER_ROWS, object_idx, object_cur, object_top);

                /* Prompt */
                prt("<dir>, ESC", 23, 0);

                /* The "current" object changed */
                if (object_old != object_idx[object_cur])
                {
                        /* Hack -- handle stuff */
                        handle_stuff();

                        /* Remember the "current" object */
                        object_old = object_idx[object_cur];
                }

                if (!column)
                {
                        Term_gotoxy(0, 6 + (grp_cur - grp_top));
                }
                else
                {
                        Term_gotoxy(max + 3, 6 + (object_cur - object_top));
                }

                ch = inkey();

                switch (ch)
                {
                        case ESCAPE:
                        {
                                flag = TRUE;
                                break;
                        }

                        default:
                        {
                                /* Move the cursor */
                                browser_cursor(ch, &column, &grp_cur, grp_cnt, &object_cur, object_cnt);
                                break;
                        }
                }
        }

        /* XXX XXX Free the "object_idx" array */
        C_KILL(object_idx, max_k_idx, int);
}


/*
 * Display known alchemy formulae.
 * A terrible, terrible hack of do_cmd_object_display().
 * Forgive me.
 */
void do_cmd_knowledge_alchemy(void)
{
        int i, len, max;
        int grp_cur, grp_top;
        int object_old, object_cur, object_top;
        int grp_cnt, grp_idx[100];
        int object_cnt;
        int *object_idx;

        int column = 0;
        bool flag;
        bool redraw;

        /* Allocate the "object_idx" array */
        C_MAKE(object_idx, max_k_idx, int);

        max = 0;
        grp_cnt = 0;

        /* Measure the label */
        len = strlen("Potions");
        max = len;

        collect_objects(1, object_idx);
        grp_idx[grp_cnt++] = 1;

        /* Terminate the list */
        grp_idx[grp_cnt] = -1;

        grp_cur = grp_top = 0;
        object_cur = object_top = 0;
        object_old = -1;

        flag = FALSE;
        redraw = TRUE;

        while (!flag)
        {
                char ch;

                if (redraw)
                {
                        clear_from(0);

                        prt("Knowledge - objects", 2, 0);
                        prt("Group", 4, 0);
                        prt("Name", 4, max + 3);
                        if (p_ptr->wizard) prt("Idx", 4, 70);
                        prt("Sym", 4, 75);

                        for (i = 0; i < 78; i++)
                        {
                                Term_putch(i, 5, TERM_WHITE, '=');
                        }

                        for (i = 0; i < BROWSER_ROWS; i++)
                        {
                                Term_putch(max + 1, 6 + i, TERM_WHITE, '|');
                        }

                        redraw = FALSE;
                }

                /* Scroll group list */
                if (grp_cur < grp_top) grp_top = grp_cur;
                if (grp_cur >= grp_top + BROWSER_ROWS) grp_top = grp_cur - BROWSER_ROWS + 1;

                /* Scroll monster list */
                if (object_cur < object_top) object_top = object_cur;
                if (object_cur >= object_top + BROWSER_ROWS) object_top = object_cur - BROWSER_ROWS + 1;

                /* Display a list of object groups */
                display_group_list(0, 6, max, BROWSER_ROWS, grp_idx, object_group_text, grp_cur, grp_top);

                /* Get a list of objects in the current group */
                object_cnt = collect_objects(grp_idx[grp_cur], object_idx);

                /* Display a list of objects in the current group */
                display_object_list(max + 3, 6, BROWSER_ROWS, object_idx, object_cur, object_top);

                /* Prompt */
                prt("<dir>, 'r' to recall, ESC", 23, 0);

                /* The "current" object changed */
                if (object_old != object_idx[object_cur])
                {
                        /* Hack -- handle stuff */
                        handle_stuff();

                        /* Remember the "current" object */
                        object_old = object_idx[object_cur];
                }

                if (!column)
                {
                        Term_gotoxy(0, 6 + (grp_cur - grp_top));
                }
                else
                {
                        Term_gotoxy(max + 3, 6 + (object_cur - object_top));
                }

                ch = inkey();

                switch (ch)
                {
                        case ESCAPE:
                        {
                                flag = TRUE;
                                break;
                        }

                        case 'R': case 'r':
                        {
                                alchemy_roff(object_idx[object_cur]);
                                msg_print(NULL);

                                redraw = TRUE;

                                break;
                        }

                        default:
                        {
                                /* Move the cursor */
                                browser_cursor(ch, &column, &grp_cur, grp_cnt, &object_cur, object_cnt);
                                break;
                        }
                }
        }

        /* XXX XXX Free the "object_idx" array */
        C_KILL(object_idx, max_k_idx, int);
}

